(** In this file we illustrate:
  *  - existential types to build heterogenous association lists;
  *  - singleton types to pattern-match on types;
  *  - equality types to encapsulate equality checking. *)

type ident = string

type _ sort =
  | String : string sort
  | Int : int sort

type binding =
  | Bind : string * 'a sort * 'a -> binding

type env = binding list

let e : env = [ Bind ("x",Int,1); Bind ("y",String,"") ]

let lookup : type a. string -> a sort -> env -> a = fun x s e ->
  let rec aux : env -> a = function
    | [] -> raise Not_found
    | Bind (y,s',v) :: env' ->
        if x <> y then aux env' else
          match s,s' with
            | Int,Int -> v
            | String,String -> v
            | Int,String | String,Int -> aux env'
  in aux e

let () = assert (1 = lookup "x" Int e)

type (_,_) eq =
  | Eq : ('a,'a) eq

let eq_sort : type a b. a sort * b sort -> (a,b) eq option = function
  | Int,Int -> Some Eq
  | String,String -> Some Eq
  | Int,String | String,Int -> None

let lookup : type a. string -> a sort -> env -> a = fun x s e ->
  let rec aux : env -> a = function
    | [] -> raise Not_found
    | Bind (y,s',v) :: env' ->
        if x <> y then aux env' else
          match eq_sort (s,s') with
            | Some Eq -> v
            | None -> aux env'
  in aux e

let () = assert (1 = lookup "x" Int e)
